@@ -1,4 +1,4 @@
-# Copyright (c) 2022, Riverbank Computing Limited
+# Copyright (c) 2023, Riverbank Computing Limited
 # All rights reserved.
 #
 # This copy of SIP is licensed for use under the terms of the SIP License
@@ -29,21 +29,24 @@ from ply import lex, yacc
 from ...bindings_configuration import get_bindings_configuration
 from ...exceptions import UserException
 
+from ..error_log import ErrorLog
+from ..instantiations import instantiate_class
 from ..python_slots import invalid_global_slot, slot_name_detail_map
+from ..scoped_name import ScopedName
 from ..specification import (AccessSpecifier, Argument, ArgumentType,
         ArrayArgument, CachedName, ClassKey, CodeBlock, Constructor,
         DocstringFormat, DocstringSignature, EnumBaseType, GILAction,
         IfaceFile, IfaceFileType, KwArgs, MappedType, Member, Module, Overload,
-        PyQtMethodSpecifier, PySlot, Qualifier, QualifierType, ScopedName,
-        Signature, Specification, Transfer, TypeHints, WrappedClass,
+        PyQtMethodSpecifier, PySlot, Qualifier, QualifierType, Signature,
+        SourceLocation, Specification, Transfer, TypeHints, WrappedClass,
         WrappedException, WrappedEnum, WrappedEnumMember)
 from ..templates import encoded_template_name, same_template_signature
-from ..utils import argument_as_str, normalised_scoped_name, same_base_type
+from ..utils import (argument_as_str, cached_name, find_iface_file,
+        normalised_scoped_name, same_base_type)
 
 from . import rules
 from . import tokens
 from .annotations import InvalidAnnotation, validate_annotation_value
-from .instantiations import instantiate_class
 
 
 class ParserManager:
@@ -52,7 +55,8 @@ class ParserManager:
     """
 
     def __init__(self, hex_version, encoding, abi_version, tags,
-            disabled_features, protected_is_public, include_dirs, strict):
+            disabled_features, protected_is_public, include_dirs, sip_module,
+            is_strict):
         """ Initialise the manager. """
 
         # Create the lexer.
@@ -73,33 +77,38 @@ class ParserManager:
         # Public state.
         self.tags = tags
 
-        self.spec = Specification()
+        self.spec = Specification(
+                tuple([int(v) for v in abi_version.split('.')]), is_strict,
+                sip_module)
+
+        # The module is initially unnamed.
+        self.modules = [self.spec.module]
 
         self.c_bindings = None
         self.code_block = None
         self.module_state = None
-        self.module_states = []
         self.paren_depth = 0
+        self.parsing_template = False
         self.parsing_virtual = False
-        self.sip_file = None
+        self.raw_sip_file = None
         self.skip_stack = [False]
 
         # Private state.
         self._hex_version = hex_version
         self._encoding = encoding
-        self._abi_version = tuple([int(v) for v in abi_version.split('.')])
         self._disabled_features = disabled_features
         self._protected_is_public = protected_is_public
         self._include_dirs = include_dirs
-        self._strict = strict
+        self._template_arg_classes = []
 
         self._scope_stack = []
-        self._errors = []
+        self._error_log = ErrorLog()
         self._file_stack = []
+        self._pending_module_state = None
         self._input = None
         self._all_sip_files = []
+        self._sip_file = None
         self._sip_files = []
-        self._name_cache = {}
 
     def complete_class(self, p, symbol, annotations, has_body):
         """ Complete the definition of the class that is the current scope, pop
@@ -122,18 +131,17 @@ class ParserManager:
         # Get the Python name and see if it is different to the C++ name.
         py_name = self.get_py_name(klass.iface_file.fq_cpp_name.base_name,
                 annotations)
-        self.check_attributes(p, symbol, py_name)
-        klass.py_name = self.cached_name(py_name)
+        klass.py_name = cached_name(self.spec, py_name)
 
         klass.no_type_hint = annotations.get('NoTypeHint', False)
 
         metatype = annotations.get('Metatype')
         if metatype is not None:
-            klass.metatype = self.cached_name(metatype)
+            klass.metatype = cached_name(self.spec, metatype)
 
         supertype = annotations.get('Supertype')
         if supertype is not None:
-            klass.supertype = self.cached_name(supertype)
+            klass.supertype = cached_name(self.spec, supertype)
 
         klass.export_derived = annotations.get('ExportDerived', False)
         klass.mixin = annotations.get('Mixin', False)
@@ -159,6 +167,10 @@ class ParserManager:
         if klass.is_opaque:
             klass.external = annotations.get('External', False)
         else:
+            # A default dtor is public.
+            if klass.dtor is None:
+                klass.dtor = AccessSpecifier.PUBLIC
+
             klass.no_default_ctors = annotations.get('NoDefaultCtors', False)
 
             # Provide a default ctor if required.
@@ -244,6 +256,9 @@ class ParserManager:
 
         self.pop_scope()
 
+        # Check the name in the current scope (ie. the class's parent scope).
+        self.check_attributes(p, symbol, py_name, ignore=klass)
+
         # Check that external classes have only been declared at the global
         # scope.
         if klass.external and self.scope is not None:
@@ -267,18 +282,49 @@ class ParserManager:
         self.push_scope(klass,
                 AccessSpecifier.PRIVATE if class_key is ClassKey.CLASS else AccessSpecifier.PUBLIC)
 
-    def find_class(self, p, symbol, iface_file_type, fq_cpp_name):
+    def disambiguate_token(self, value, keywords):
+        """ Disambiguate a token by inspecting its value. """
+
+        # This seems to be needed because it's not possible to get lex() to do
+        # it.  The problem seems to be that you can't control the order in
+        # which lex() applies its regular expressions despite what the
+        # documentation says.  It seems that tokens that are specific to a
+        # state are always added after everything else no matter where they
+        # appear in the file.
+
+        if value in keywords:
+            token_type = value
+        elif value == '...':
+            token_type = 'ELLIPSIS'
+        elif value.startswith('.'):
+            token_type = 'FILE_PATH'
+        else:
+            for marker in ('/', '..', '-'):
+                if marker in value:
+                    token_type = 'FILE_PATH'
+                    break
+            else:
+                if '.' in value:
+                    token_type = 'DOTTED_NAME'
+                else:
+                    token_type = 'NAME'
+
+        return token_type
+
+    def find_class(self, p, symbol, iface_file_type, fq_cpp_name,
+            tmpl_arg=False):
         """ Return a WrappedClass object for a C++ name creating it if
         necessary.
         """
 
         return self._find_class_with_iface_file(
-                self.find_iface_file(p, symbol, fq_cpp_name, iface_file_type))
+                self.find_iface_file(p, symbol, fq_cpp_name, iface_file_type),
+                tmpl_arg=tmpl_arg)
 
     def find_exception(self, p, symbol, fq_cpp_name, raise_code=None):
         """ Find an exception, optionally creating a new one. """
 
-        fq_cpp_name = normalised_scoped_name(fq_cpp_name, self.scope)
+        # Note that we don't normalise the name.
         iface_file = self.find_iface_file(p, symbol, fq_cpp_name,
                 IfaceFileType.EXCEPTION)
 
@@ -378,7 +424,7 @@ class ParserManager:
 
         return klass
 
-    def _find_class_with_iface_file(self, iface_file):
+    def _find_class_with_iface_file(self, iface_file, tmpl_arg=False):
         """ Return a WrappedClass object for an interface file creating it if
         necessary.
         """
@@ -386,11 +432,20 @@ class ParserManager:
         # See if it already exists.
         for klass in self.spec.classes:
             if klass.iface_file is iface_file:
+                if not self.parsing_template:
+                    try:
+                        self._template_arg_classes.remove(klass)
+                    except ValueError:
+                        pass
+
                 return klass
 
         # Create a new one.
         klass = WrappedClass(iface_file,
-                self.cached_name(iface_file.fq_cpp_name.base_name), None)
+                cached_name(self.spec, iface_file.fq_cpp_name.base_name), None)
+
+        if tmpl_arg:
+            self._template_arg_classes.append(klass)
 
         # Use the same ordering as the old parser.
         self.spec.classes.insert(0, klass)
@@ -428,7 +483,7 @@ class ParserManager:
                 result=Argument(ArgumentType.VOID))
         self._check_ellipsis(p, symbol, py_signature)
 
-        # Configure the destructor.
+        # Configure the constructor.
         ctor = Constructor(access_specifier, py_signature)
 
         if annotations.get("NoDerived", False):
@@ -477,11 +532,6 @@ class ParserManager:
 
         ctor.transfer = self.get_transfer(p, symbol, annotations)
 
-        # /Transfer/ arguments need the wrapper.
-        for ad in py_signature.args:
-            if ad.transfer is Transfer.TRANSFER:
-                ad.get_wrapper = True
-
         scope.ctors.append(ctor)
 
     def add_dtor(self, p, symbol, name, annotations, *, exceptions,
@@ -550,7 +600,7 @@ class ParserManager:
         base_type = EnumBaseType.ENUM
 
         if base_type_s is not None:
-            if self._abi_version < (13, 0):
+            if self.spec.abi_version < (13, 0):
                 self.parser_error(p, symbol,
                         "/BaseType/ is only supported for ABI v13.0 and later")
 
@@ -575,8 +625,9 @@ class ParserManager:
             py_name = None
         else:
             fq_cpp_name = normalised_scoped_name(cpp_name, self.scope)
-            cached_fq_cpp_name = self.cached_name(str(fq_cpp_name))
-            py_name = self.cached_name(self.get_py_name(cpp_name, annotations))
+            cached_fq_cpp_name = cached_name(self.spec, str(fq_cpp_name))
+            py_name = cached_name(self.spec,
+                    self.get_py_name(cpp_name, annotations))
 
             self.check_attributes(p, symbol, py_name)
 
@@ -668,9 +719,7 @@ class ParserManager:
         overload.throw_args = exceptions
         overload.virtual_call_code = virtual_call_code
         overload.virtual_catcher_code = virtual_catcher_code
-
-        overload.source_file = self.sip_file
-        overload.source_line = p.lineno(symbol)
+        overload.source_location = self.get_source_location(p, symbol)
 
         # See if the function is a non-lazy method.  These are methods that
         # Python expects to see defined in the type before any instance of the
@@ -767,6 +816,10 @@ class ParserManager:
                 len_method_code = CodeBlock("Auto-generated",
                         text='            sipRes = (Py_ssize_t)sipCpp->{0}();\n'.format(cpp_name))
 
+                # Note that we can't currently use any %MethodCode because it
+                # does too much (ie. setting sipRes).
+                self.scope.len_cpp_name = cpp_name
+
             len_py_signature = Signature(result=Argument(ArgumentType.SSIZE))
 
             self._add_auto_slot(p, symbol, annotations, '__len__',
@@ -804,7 +857,7 @@ class ParserManager:
             fq_cpp_name = ''
 
         iface_file = self.find_iface_file(p, symbol, fq_cpp_name,
-                IfaceFileType.MAPPED_TYPE, cpp_type)
+                IfaceFileType.MAPPED_TYPE, cpp_type=cpp_type)
 
         # Check it hasn't already been defined.
         for mtd in self.spec.mapped_types:
@@ -821,10 +874,11 @@ class ParserManager:
         # Create a new mapped type.
         mapped_type = MappedType(iface_file, cpp_type)
 
-        mapped_type.cpp_name = self.cached_name(argument_as_str(cpp_type))
+        mapped_type.cpp_name = cached_name(self.spec,
+                argument_as_str(cpp_type))
 
         if cpp_name is not None:
-            mapped_type.py_name = self.cached_name(
+            mapped_type.py_name = cached_name(self.spec,
                     annotations.get('PyName', cpp_name))
 
         self.annotate_mapped_type(p, symbol, mapped_type, annotations)
@@ -869,7 +923,7 @@ class ParserManager:
     def add_typedef(self, p, symbol, typedef):
         """ Add a typedef to the current scope. """
 
-        if self._strict:
+        if self.spec.is_strict:
             for td in self.spec.typedefs:
                 if td.fq_cpp_name == typedef.fq_cpp_name:
                     self.parser_error(p, symbol,
@@ -884,9 +938,18 @@ class ParserManager:
         """ Apply annotations to a mapped type. """
 
         mapped_type.handles_none = annotations.get('AllowNone', False)
+        mapped_type.no_assignment_operator = annotations.get(
+                'NoAssignmentOperator', False)
+        mapped_type.no_copy_ctor = annotations.get('NoCopyCtor', False)
+        mapped_type.no_default_ctor = annotations.get('NoDefaultCtor', False)
         mapped_type.no_release = annotations.get('NoRelease', False)
         mapped_type.type_hints = self.get_type_hints(p, symbol, annotations)
 
+        if mapped_type.no_release:
+            mapped_type.no_assignment_operator = True
+            mapped_type.no_copy_ctor = True
+            mapped_type.no_default_ctor = True
+
         pyqt_flags = self._get_plugin_annotation(p, symbol, annotations,
                 'PyQtFlags', 'PyQt6')
         if pyqt_flags is not None:
@@ -954,32 +1017,14 @@ class ParserManager:
                         "{0} is not a valid {1} annotation".format(name,
                                 context))
 
-    def cached_name(self, name):
-        """ Add a name to the cache if necessary and return the cached name.
-        """
-
-        # Get the line of the cache for the length of this name creating it if
-        # necessary.
-        line = self._name_cache.setdefault(len(name), [])
-
-        # See if the name has already been cached.
-        for nd in line:
-            if nd.name == name:
-                return nd
-
-        # Create a new entry.
-        nd = CachedName(name)
-        line.append(nd)
-
-        return nd
-
-    def check_attributes(self, p, symbol, py_name, is_function=False):
-        """ Check that a Python name will not class with another object in the
+    def check_attributes(self, p, symbol, py_name, is_function=False,
+            ignore=None):
+        """ Check that a Python name will not clash with another object in the
         same Python scope.
         """
 
         # We don't do any check for a non-strict parse.
-        if not self._strict:
+        if not self.spec.is_strict:
             return
 
         # Report a name clash with something.
@@ -1037,6 +1082,11 @@ class ParserManager:
             if cd.scope is not self.scope:
                 continue
 
+            # A class will have already been added to the scope and this will
+            # tell us to ignore it.
+            if cd is ignore:
+                continue
+
             if cd.external:
                 continue
 
@@ -1155,19 +1205,14 @@ class ParserManager:
         things so that it appears like the former.
         """
 
-        sip_file, input, lineno, lexpos, module_state = self._file_stack.pop()
+        sip_file, raw_sip_file, input, lineno, lexpos, module_state = self._file_stack.pop()
 
         if module_state is None:
+            self._import_module(self._sip_file)
             module_state = self.module_state
 
-            # The current file was %Included so create a new Module and
-            # ModuleState as if it had been %Imported.
-            module = Module()
-            self.spec.modules.append(module)
-            self.module_state = ModuleState(module, self.sip_file, self)
-
         self._file_stack.append(
-                (sip_file, input, lineno, lexpos, module_state))
+                (sip_file, raw_sip_file, input, lineno, lexpos, module_state))
 
     def evaluate_feature_or_platform(self, p, symbol, name=None,
             inverted=False):
@@ -1184,7 +1229,7 @@ class ParserManager:
             value = qual.name not in self._disabled_features
         elif qual.type is QualifierType.PLATFORM:
             # The platform is always ignored in non-strict mode.
-            if not self._strict:
+            if not self.spec.is_strict:
                 return True
 
             value = qual.name in self.tags
@@ -1273,7 +1318,7 @@ class ParserManager:
     def find_qualifier(self, p, symbol, name, required=True):
         """ Return a Qualifier or None if one doesn't exist. """
 
-        for module in self.spec.modules:
+        for module in self.modules:
             for qual in module.qualifiers:
                 if qual.name == name:
                     return qual
@@ -1396,14 +1441,15 @@ class ParserManager:
 
             return None
 
-        # Check that type hints haven't been suppressed.
-        if annotations.get('NoTypeHint') is not None and (th_in or th_out or th_value):
-            self.parser_error(p, symbol,
-                    "'NoTypeHint' cannot be specified with a type hint")
+        if th_in is not None or th_out is not None or th_value is not None:
+            # Check that type hints haven't been suppressed.
+            if annotations.get('NoTypeHint') is not None:
+                self.parser_error(p, symbol,
+                        "'NoTypeHint' cannot be specified with a type hint")
 
-            return None
+            return TypeHints(th_in, th_out, th_value)
 
-        return TypeHints(th_in, th_out, th_value)
+        return None
 
     @property
     def in_main_module(self):
@@ -1411,9 +1457,7 @@ class ParserManager:
         code will be generated for.
         """
 
-        module = self.module_state.module
-
-        return module is self.spec.modules[0] or module.composite is not None
+        return self.module_state.module is self.spec.module
 
     def instantiate_class_template(self, p, symbol, fq_cpp_name, template,
             py_name, no_type_name, docstring):
@@ -1423,7 +1467,7 @@ class ParserManager:
 
         # Look for an appropriate class template.
         for tmpl_names, proto_class in self.class_templates:
-            if proto_class.iface_file.fq_cpp_name.matches(template.cpp_name, self.scope) and same_template_signature(tmpl_names, template.types):
+            if proto_class.iface_file.fq_cpp_name.matches(template.cpp_name, scope=self.scope) and same_template_signature(tmpl_names, template.types):
                 break
         else:
             # There was no class template to instantiate.
@@ -1437,76 +1481,100 @@ class ParserManager:
     def lexer_error(self, t, text):
         """ Record an error caused by a token. """
 
-        self._errors.append(
-                self._format_error(self.sip_file, t.lineno, t.lexpos, text))
+        self._error_log.log(text,
+                SourceLocation(self._sip_file, line=t.lineno,
+                        column=self._get_column_from_lexpos(t.lexpos)))
 
     def parse(self, sip_file):
-        """ Parse a .sip file and return a Specification object and a list of
-        the .sip files that specify the module to be generated.  A
-        UserException is raised if there was an error.
+        """ Parse a .sip file and return a 3-tuple of a Specification object, a
+        list of Module objects and a list of the .sip files that specify the
+        module to be generated.  A UserException is raised if there was an
+        error.
         """
 
+        # Note that the retention of the 'raw' filename, ie. that which was
+        # specified by the user is only done so that generated '#line'
+        # directives match those from older versions of SIP.
+
+        raw_sip_file = sip_file
         sip_file = os.path.abspath(sip_file)
 
-        self.module_state = ModuleState(self.spec.modules[0], sip_file, self)
+        self.module_state = ModuleState(self.spec.module, sip_file)
 
         try:
-            self._parser.parse(self._read(sip_file), lexer=self._lexer,
-                    tracking=True)
+            self._parser.parse(self._read(sip_file, raw_sip_file),
+                    lexer=self._lexer, tracking=True)
         except UnexpectedEOF:
             self._unexpected_eof_error()
 
         self._handle_eom()
-        self._report_errors()
+        self._error_log.as_exception()
 
         self.spec.c_bindings = bool(self.c_bindings)
 
         self.spec.typedefs.sort(key=lambda k: k.fq_cpp_name)
         self.spec.variables.sort(key=lambda k: k.py_name.name)
 
-        # Remove all template classes.
-        for _, klass in self.class_templates:
+        # Remove all template classes and anything they contain.
+        template_classes = [k for _, k in self.class_templates]
+
+        for enum in list(self.spec.enums):
+            if enum.scope in template_classes:
+                spec.spec.enums.remove(enum)
+
+        for typedef in list(self.spec.typedefs):
+            if typedef.scope in template_classes:
+                spec.spec.typedefs.remove(typedef)
+
+        for variable in list(self.spec.variables):
+            if variable.scope in template_classes:
+                spec.spec.variables.remove(variable)
+
+        for klass in template_classes:
             self.spec.classes.remove(klass)
             self.spec.iface_files.remove(klass.iface_file)
 
-        # Create the name cache as seen by the rest of the code.
-        for k in sorted(self._name_cache.keys(), reverse=True):
-            self.spec.name_cache.extend(
-                    sorted(self._name_cache[k], key=lambda k: k.name))
+        # Remove all classes that are only template arguments.
+        for klass in self._template_arg_classes:
+            self.spec.classes.remove(klass)
 
-        return self.spec, self._sip_files
+        return self.spec, self.modules, self._sip_files
 
     def parser_error(self, p, symbol, text):
         """ Record an error caused by a symbol in a production. """
 
-        self._errors.append(
-                self._format_error(self.sip_file, p.lineno(symbol),
-                        p.lexpos(symbol), text))
+        self._error_log.log(text, self.get_source_location(p, symbol))
 
     def pop_file(self):
         """ Restore the current .sip file from the stack and make it current.
         An IndexError is raised if the stack is empty.
         """
 
-        # Restore the state of the previous .sip file.
-        sip_file, self._input, self._lexer.lineno, lexpos, old_module_state = self._file_stack.pop()
+        # Restore the state of the previous .sip file.  Note that we don't
+        # restore the module state until after the EOF has been seen.
+        self._sip_file, self.raw_sip_file, self._input, self._lexer.lineno, lexpos, self._pending_module_state = self._file_stack.pop()
+
         self._lexer.input(self._input)
         self._lexer.lexpos = lexpos
 
-        if old_module_state is not None:
-            self._handle_eom()
+    def pop_module_state(self):
+        """ Restore the current module state. """
 
-            # Inherit any default encoding.
-            if old_module_state.default_encoding is None:
-                old_module_state.default_encoding = self.module_state.default_encoding
+        if self._pending_module_state is None:
+            return
 
-            # Inherit any call_super_init.
-            if old_module_state.call_super_init is None:
-                old_module_state.call_super_init = self.module_state.call_super_init
+        self._handle_eom()
 
-            self.module_state = old_module_state
+        # Inherit any default encoding.
+        if self._pending_module_state.default_encoding is None:
+            self._pending_module_state.default_encoding = self.module_state.default_encoding
 
-        self.sip_file = sip_file
+        # Inherit any call_super_init.
+        if self._pending_module_state.call_super_init is None:
+            self._pending_module_state.call_super_init = self.module_state.call_super_init
+
+        self.module_state = self._pending_module_state
+        self._pending_module_state = None
 
     def pop_scope(self):
         """ Pop the current scope. """
@@ -1523,6 +1591,8 @@ class ParserManager:
         if sip_file is None:
             sip_file = p[symbol]
 
+        raw_sip_file = sip_file
+
         # Make the name platform-native.
         sip_file = sip_file.replace('/', os.sep)
 
@@ -1535,7 +1605,7 @@ class ParserManager:
             # If the name is relative then check the directory containing the
             # current file and any include directories.
             if not os.path.isabs(sip_file):
-                inc_dirs = [os.path.dirname(self.sip_file)]
+                inc_dirs = [os.path.dirname(self._sip_file)]
                 inc_dirs.extend(self._include_dirs)
 
                 for inc_dir in inc_dirs:
@@ -1547,11 +1617,12 @@ class ParserManager:
             if found is None:
                 if not optional:
                     self.parser_error(p, symbol,
-                            "'{0}' could not be found".format(sip_file))
+                            "'{0}' could not be found".format(raw_sip_file))
 
                 return
 
-            sip_file = found
+            # For historic reasons we keep the absolute name for the raw name.
+            raw_sip_file = sip_file = found
 
         sip_file = os.path.abspath(sip_file)
 
@@ -1562,55 +1633,26 @@ class ParserManager:
                         "'{0}' is being read recursively".format(sip_file))
                 return
 
+        # Ignore the file if we have already read it.
+        if sip_file in self._all_sip_files:
+            return
+
         if new_module:
-            importing_from = self.module_state.module
-
-            # Create a new module if it has not already been defined.
-            for module_state in self.module_states:
-                if module_state.sip_file == sip_file:
-                    module = module_state.module
-                    break
-            else:
-                module = Module()
-                self.spec.modules.append(module)
-
-                module.default_exception = self.module_state.module.default_exception
-                old_module_state = self.module_state
-                self.module_state = ModuleState(module, sip_file, self)
-
-                # Get the configuration of the new module.
-                mod_tags, mod_disabled = get_bindings_configuration(
-                        self._abi_version[0], sip_file, self._include_dirs)
-
-                for tag in mod_tags:
-                    if tag not in self.tags:
-                        self.tags.append(tag)
-
-                for feature in mod_disabled:
-                    if feature not in self._disabled_features:
-                        self._disabled_features.append(feature)
-
-            # Add the new import unless it has already been imported.
-            if module not in importing_from.imports:
-                importing_from.imports.insert(0, module)
+            old_module_state = self.module_state
+            self._import_module(sip_file)
         else:
             # This means that the file was %Included rather than %Imported.
             old_module_state = None
 
-        # Ignore the file if we have already read it. This replicates the
-        # behaviour of the old parser but shouldn't be required for a well
-        # specified project.
-        if sip_file in self._all_sip_files:
-            return
-
         # Save the state of the current .sip file.
         self._file_stack.append(
-                (self.sip_file, self._input, self._lexer.lineno,
-                        self._lexer.lexpos, old_module_state))
+                (self._sip_file, self.raw_sip_file, self._input,
+                        self._lexer.lineno, self._lexer.lexpos,
+                        old_module_state))
 
         # Make the new one current and give it's content to the lexer.
         self._lexer.lineno = 1
-        self._lexer.input(self._read(sip_file))
+        self._lexer.input(self._read(sip_file, raw_sip_file))
 
     def push_scope(self, scope, access_specifier=None):
         """ Push a new scope. """
@@ -1744,7 +1786,7 @@ class ParserManager:
     def validate_mapped_type(self, p, symbol, mapped_type):
         """ Validate a completed mapped type. """
 
-        if self._abi_version >= (13, 0):
+        if self.spec.abi_version >= (13, 0):
             convert_to_us = mapped_type.convert_to_type_code is not None and 'sipUserState' in mapped_type.convert_to_type_code.text
 
             release_us = mapped_type.release_code is not None and 'sipUserState' in mapped_type.release_code.text
@@ -1820,6 +1862,11 @@ class ParserManager:
 
         for arg in signature.args:
             if arg.type is ArgumentType.ELLIPSIS:
+                # Give the argument the standard name so that it appears in
+                # docstrings and type hints.  It is never used in generated
+                # code so there is no need to add it to the module cache.
+                arg.name = CachedName('*args')
+
                 if seen_ellipsis:
                     self.parser_error(p, symbol,
                             "'...' may only be specified once")
@@ -1837,67 +1884,12 @@ class ParserManager:
         creating it if necessary.
         """
 
-        # See if the name is already used.
-        for iff in self.spec.iface_files:
-            if not iff.fq_cpp_name.matches(fq_cpp_name, self.scope):
-                continue
+        def error_logger(text):
+            self.parser_error(p, symbol, text)
 
-            # They must be the same type except that we allow a class if we
-            # want an exception.  This is because we allow classes to be used
-            # before they are defined.
-            if iff.type is not iface_file_type:
-                if iface_file_type is not IfaceFileType.EXCEPTION or iff.type is not IfaceFileType.CLASS:
-                    self.parser_error(p, symbol,
-                            "a class, exception, namespace or mapped type has already been defined with the same name")
-
-            # Ignore an external class declared in another module.
-            if iface_file_type is IfaceFileType.CLASS and iff.module is not None and iff.module is not self.module_state.module:
-                for cd in self.spec.classes:
-                    if cd.iface_file is iff:
-                        external = cd.external
-                        break
-                else:
-                    external = False
-
-                if external:
-                    continue
-
-            # If this is a mapped type with the same name defined in a
-            # different module, then check that this type isn't the same as any
-            # of the mapped types defined in that module.
-            if iface_file_type == IfaceFileType.MAPPED_TYPE and iff.module is not self.module_state.module:
-                for mtd in self.spec.mapped_types:
-                    if mtd.iface_file is not iff:
-                        continue
-
-                    if cpp_type.type is not ArgumentType.TEMPLATE or \
-                        mtd.type.type is not ArgumentType.TEMPLATE or \
-                        same_base_type(cpp_type, mtd.type):
-                        self.parser_error(p, symbol,
-                                "the mapped type has already been defined in another module")
-
-                # If we got here then we have a mapped type based on an
-                # existing template, but with unique parameters.  We don't want
-                # to use interface files from other modules, so skip this one.
-                continue
-
-            # Ignore a namespace defined in another module.
-            if iface_file_type is IfaceFileType.NAMESPACE and iff.module is not self.module_state.module:
-                continue
-
-            return iff
-
-        # Create a new interface file.
-        fq_cpp_name = normalised_scoped_name(fq_cpp_name, self.scope)
-
-        iff = IfaceFile(iface_file_type,
-                cpp_name=self.cached_name(str(fq_cpp_name)),
-                fq_cpp_name=fq_cpp_name)
-
-        # Use the same ordering as the old parser.
-        self.spec.iface_files.insert(0, iff)
-
-        return iff
+        return find_iface_file(self.spec, self.module_state.module,
+                fq_cpp_name, iface_file_type, error_logger, cpp_type=cpp_type,
+                scope=self.scope)
 
     def _find_member(self, p, symbol, py_name, args, annotations, method_code):
         """ Return (creating if necessary) the member for a Python name. """
@@ -1934,6 +1926,9 @@ class ParserManager:
 
             # __delattr__ is implemented as __setattr__.
             if py_slot is PySlot.DELATTR:
+                if self.in_main_module:
+                    cached_name(self.spec, py_name).used = True
+
                 py_slot = PySlot.SETATTR
                 py_name = '__setattr__'
 
@@ -1966,7 +1961,7 @@ class ParserManager:
                 break
         else:
             member = Member(self.module_state.module,
-                    self.cached_name(py_name))
+                    cached_name(self.spec, py_name))
             member.namespace_iface_file = namespace_iface_file
             member.no_arg_parser = no_arg_parser
             member.py_slot = py_slot
@@ -1994,13 +1989,14 @@ class ParserManager:
 
         return qual
 
-    def _format_error(self, sip_file, line, lexpos, text):
-        """ Return a formated error message. """
+    def _get_column_from_lexpos(self, lexpos):
+        """ Return the column within the current line corresponding to a lexer
+        position.
+        """
 
         line_start = self._input.rfind('\n', 0, lexpos) + 1
-        col = lexpos - line_start + 1
 
-        return "{0}: line {1} column {2}: {3}".format(sip_file, line, col, text)
+        return lexpos - line_start + 1
 
     def _get_gil_action(self, p, symbol, annotations):
         """ Return an appropriate GILAction according to the annotations. """
@@ -2072,6 +2068,12 @@ class ParserManager:
 
         return anno
 
+    def get_source_location(self, p, symbol):
+        """ Return a SourceLocation object for a symbol. """
+
+        return SourceLocation(self._sip_file, line=p.lineno(symbol),
+                column=self._get_column_from_lexpos(p.lexpos(symbol)))
+
     def _handle_eom(self):
         """ Check that the current module is complete. """
 
@@ -2079,14 +2081,41 @@ class ParserManager:
         module = module_state.module
 
         if module.fq_py_name is None:
-            self._errors.append(
-                    "{0}: %Module has not been specified".format(
-                            self.sip_file))
+            self._error_log.log("%Module has not been specified",
+                    SourceLocation(self._sip_file))
 
         # call_super_init defaults to False if it wasn't specified.
         module.call_super_init = bool(module_state.call_super_init)
 
-    def _read(self, sip_file):
+    def _import_module(self, sip_file):
+        """ Create a new Module object and corresponding ModuleState object for
+        a .sip file and make it current.
+        """
+
+        importing_from = self.module_state.module
+
+        module = Module()
+        self.modules.append(module)
+
+        module.default_exception = self.module_state.module.default_exception
+        self.module_state = ModuleState(module, sip_file)
+
+        # Get the configuration of the new module.
+        mod_tags, mod_disabled = get_bindings_configuration(
+                self.spec.abi_version[0], sip_file, self._include_dirs)
+
+        for tag in mod_tags:
+            if tag not in self.tags:
+                self.tags.append(tag)
+
+        for feature in mod_disabled:
+            if feature not in self._disabled_features:
+                self._disabled_features.append(feature)
+
+        # Add the new import.
+        importing_from.imports.append(module)
+
+    def _read(self, sip_file, raw_sip_file):
         """ Return the contents of the current .sip file. """
 
         try:
@@ -2100,7 +2129,8 @@ class ParserManager:
                             sip_file, self._encoding),
                     detail=str(e))
 
-        self.sip_file = sip_file
+        self.raw_sip_file = raw_sip_file
+        self._sip_file = sip_file
         self._all_sip_files.append(sip_file)
 
         if self.in_main_module:
@@ -2108,23 +2138,17 @@ class ParserManager:
 
         return self._input
 
-    def _report_errors(self):
-        """ Raise an exception if there any errors to report. """
-
-        if self._errors:
-            raise UserException('\n'.join(self._errors))
-
     def _unexpected_eof_error(self):
         """ Record an error caused by an unexpected EOF. """
 
-        self._errors.append(
-                "{0}: unexpected end of file".format(self.sip_file))
+        self._error_log.log("unexpected end of file",
+                SourceLocation(self._sip_file))
 
 
 class ModuleState:
     """ Encapsulate the parser-related state for a module. """
 
-    def __init__(self, module, sip_file, pm):
+    def __init__(self, module, sip_file):
         """ Initialise the state. """
 
         self.module = module
@@ -2137,8 +2161,6 @@ class ModuleState:
         self.kw_args = KwArgs.NONE
         self.nr_timelines = 0
 
-        pm.module_states.append(self)
-
 
 class ScopeState:
     """ Encapsulate the parser-related state for a scope. """
